/*
 * Fick - Version 0.3
 *
 *  Copyright (C) 2005  Danilo Abbasciano
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Bug Report: <danilo.abbasciano@gmail.com>
 *
 */

#include "header.h"

int main(int argc, char **argv) {
  SDL_Event event;
  SDL_Surface *screen;
  char *file_out = NULL;
  int Dx = 0, Dy = 0, w = 0, h = 0;
  float Dt = 0;
  float **A = NULL, *b;
  int option_index, c, steps = -1, delay = TIME_INTERVAL,
    multicolor = 0;
  Matrice D, C;
  char t[20];

  while (1) {
    option_index = 0;
    if ((c = getopt_long(argc, argv, "hvmqex:y:t:s:o:d:W:H:f", long_options,
                         &option_index)) == -1)
      break;
    
    switch (c) {
    case 0:
      break;
    case 'h':
      printf("fick version 0.3\n"
             "Usage: %s [OPTIONS] ImageDistribition [ImageDiffusion]\n\n"
             "This program supports images from the following formats:\n"
             " TGA, BMP, PNM (include: pnm, pbm, pgm, ppm), XPM, XCF, PCX,\n"
             " GIF, JPG, JPEG, TIF, TIFF, LBM, IFF, PNG\n\n"
             "OPTIONS:\n"
             "  -f --fullscreen\t\t: full screen mode\n"
             "  -e --explicit\t\t\t: use explicit method\n"
	     "  -m --multicolor\t\t: 2d view with colors\n"
             "  --3d\t\t\t\t: 3d mode\n"
             "  -x <arg> --dx <arg>\t\t: set Dx\n"
             "  -y <arg> --dy <arg>\t\t: set Dy\n"
             "  -t <arg> --dt <arg>\t\t: set Dt\n"
             "  -s <n> --steps <n>\t\t: the program stops after <n> steps\n"
             "  -d <s> --delay <s>\t\t: ms between two image (default 250)\n"
             "  -W <n> --Width <n>\t\t: width of the surface\n"
             "  -H <n> --Height <n>\t\t: height of the surface\n"
             "  -o <file> --output <file>\t: save results as a BMP <file>\n"
             "  -v --version\t\t\t: print version info and exit\n"
             "  -h --help\t\t\t: display this help and exit\n\n"
             "Report bugs to <danilo.abbasciano@gmail.com>\n", argv[0]);
      return 1;
    case 'v':
      printf("fick-0.3\n\n"
             "Written by Danilo Abbasciano.\n\n"
             
             "This is free software; see the source for copying conditions."
             "There is NO\n"
             "warranty; not even for MERCHANTABILITY or FITNESS FOR A"
             "PARTICULAR PURPOSE.\n");
      return 1;
    case 'm': multicolor = 1;
      break;
    case 'e': implicit = 0;
      break;
    case 'x': sscanf(optarg, "%d", &Dx);
      break;
    case 'y': sscanf(optarg, "%d", &Dy);
      break;
    case 't': sscanf(optarg, "%f", &Dt);
      break;
    case 's': sscanf(optarg, "%d", &steps);
      break;
    case 'o': file_out = optarg;
      break;
    case 'd': sscanf(optarg, "%d", &delay);
      break;
    case 'W': sscanf(optarg, "%d", &w);
      break;
    case 'H': sscanf(optarg, "%d", &h);
      break;
    case 'f': full_screen = SDL_FULLSCREEN;
      break;
    case '?': default: 
      return 1;
    }
  }

  if (argc - optind != 1 && argc - optind != 2) {
    fprintf(stdout, "error: Try `%s --help' for more information.\n", argv[0]);
    return 1;
  }
 
  /* Verify the stability restriction */
  if (!implicit && Dt > Dx*Dx*Dy*Dy / (2.0 * (Dx*Dx + Dy*Dy))) {
    Dt = Dx*Dx*Dy*Dy / (2.0 * (Dx*Dx + Dy*Dy));
    printf("For rispect the stability restriction must be used Dt = %f\n", Dt);
  }

  /* inizializing SDL */
  if (SDL_Init (SDL_INIT_VIDEO | SDL_INIT_TIMER) < 0) {
    fprintf(stderr, "Impossibile inizializzare SDL: %s\n", SDL_GetError() );
    exit(1);
  }
  atexit(SDL_Quit);

  while (Dx <= 0) {
    printf("Dx > ");
    scanf("%d", &Dx);
  }
  while (Dy <= 0) {
    printf("Dy > ");
    scanf("%d", &Dy);
  }
  while (Dt <= 0) {
    printf("Dt > ");
    scanf("%f", &Dt);
  }

  if (w == 0) w = Dx;
  if (h == 0) h = Dy;

  /* Load the D and C matrix with the color R(Red) of diffuzion image file
     and color B(Blue) of distribution image file. One element of matrix are
     the mean of a partition of image.*/

  C.color = B;
  image2matrix(loadImage(argv[optind]), Dx, Dy, &C);

  D.color = R;
  image2matrix(loadImage(argv[optind + 1]), Dx, Dy, &D);
  
  C.r = D.r = Min(C.r, D.r);
  C.c = D.c = Min(C.c, D.c);

  SDL_WM_SetCaption("fick", NULL);
  
  printf("Wait...\n");

  if (implicit) {
    /* making a values on A matrix */
    A = (float **)coefficient_matrix(Dx, Dy, Dt, &D);

    /* decompose the matrix A in two matrix(L and U). L and U are stored in A*/
    LUdecomposition(A, D.r*D.c);
  }
    
  if (mode3d) {
    
    screen = SDL_SetVideoMode(2*BORDER + COS30*(w * (C.r - 1) + h*(C.c - 1)),
                              2*BORDER + SIN30*(w * (C.r - 1) + h*(C.c - 1)) + 255,
                              BPP,
                              SDL_HWSURFACE | SDL_ANYFORMAT | full_screen );

    background.w = screen -> w;
    background.h = screen -> h;
    
    matrix2image3d(screen, w, h, &C);
  } else {
    screen = SDL_SetVideoMode(C.c * w, C.r * h, BPP,
                              SDL_HWSURFACE | SDL_ANYFORMAT | full_screen );
    
    matrix2image(screen, w, h, &C, multicolor);
  }
  
  SDL_UpdateRect(screen, 0, 0, screen -> w, screen -> h);
  
  printf("Start simulation\n");

  while (steps != 0) {
    /* calculate_b resolv: b = (2Id + A)Ct end return the b vector. */
    b = (float *)calculate_b(Dx, Dy, Dt, &D, &C);

    if (implicit)

      /* LUsolve delete the concentration at time t and 
         resolv the system L*U*C(t+1) = b,
	 return the results in a vector.  */
      LUsolve(A, b, D.r*D.c, C.m);
    else { /* explicit mode */
      for (c = 0; c < D.r*D.c; c++) {
        if (b[c] + 0.5 > 255) C.m[c] = 255;
        else C.m[c] = (Uint8)(b[c] + 0.5);
      }
      free(b);
    }

    if (mode3d) matrix2image3d(screen, w, h, &C);
    else matrix2image(screen, w, h, &C, multicolor);

    SDL_Delay(delay);
    /* showImage  */
    SDL_UpdateRect(screen, 0, 0, screen -> w, screen -> h);

    while (SDL_PollEvent(&event) == 1) {
      switch (event.type) {
      case SDL_KEYDOWN:
        if(event.key.keysym.sym == SDLK_f)
          SDL_WM_ToggleFullScreen(screen);

        if (event.key.keysym.sym == SDLK_ESCAPE ||
            event.key.keysym.sym == SDLK_q)
          stop(screen, file_out, C.m, D.m, A, D.r*D.c);

        if (event.key.keysym.sym == SDLK_PLUS)
          Dt++;
        
        if (event.key.keysym.sym == SDLK_MINUS)
          if (Dt > 0) Dt--;

        if (event.key.keysym.sym == SDLK_o) {
           if (SDL_SaveBMP(screen, FILE_OUT) == -1)
             fprintf(stderr, "error\n");
        }
        
        break;
      case SDL_QUIT:
        stop(screen, file_out, C.m, D.m, A, D.r*D.c);
      }
    }
    steps = steps - 1;
  }
  stop(screen, file_out, C.m, D.m, A, D.r*D.c);
  return 0;
}
