#include "header.h"

Uint32 getpixel(SDL_Surface *surface, int x, int y) {
    int bpp = surface->format->BytesPerPixel;
    /* Here p is the address to the pixel we want to retrieve */
    Uint8 *p = (Uint8 *)surface->pixels + y * surface->pitch + x * bpp;

    switch(bpp) {
    case 1:
        return *p;

    case 2:
        return *(Uint16 *)p;

    case 3:
        if(SDL_BYTEORDER == SDL_BIG_ENDIAN)
            return p[0] << 16 | p[1] << 8 | p[2];
        else
            return p[0] | p[1] << 8 | p[2] << 16;

    case 4:
        return *(Uint32 *)p;

    default:
        return 0;
    }
}

void putpixel(SDL_Surface *s, int x, int y, Uint32 pixel) {
  int bpp = s -> format -> BytesPerPixel;
  Uint8 *p = (Uint8 *)s -> pixels + y * s -> pitch + x * bpp;

  switch (bpp) {
  case 1: *p = pixel;
    break;
  case 2: *(Uint16 *)p = pixel;
    break;
  case 3:
    if (SDL_BYTEORDER == SDL_BIG_ENDIAN) {
      p[0] = (pixel >> 16) & 0xff;
      p[2] = pixel & 0xff;
    }else {
      p[0] = pixel & 0xff;
      p[2] = (pixel >> 16) & 0xff;
    }
    p[1] = (pixel >> 8) & 0xff;
    break;
  case 4: *(Uint32 *)p = pixel;
  }
}

Uint8 getColor(SDL_Surface *s, int x, int y, int color) {
  Uint8 v[3];
  SDL_GetRGB(getpixel(s, x, y), s -> format, &v[R], &v[G], &v[B]);
  return v[color];
}

/* Lock the screen for direct access to the pixels */
void lockScreen(SDL_Surface *s) {
  if (SDL_MUSTLOCK(s))
    if ( SDL_LockSurface(s) < 0 ) {
      fprintf(stderr, "Can't lock screen: %s\n", SDL_GetError());
      exit (1);
    }
}

void unlockScreen(SDL_Surface *s) {
  if (SDL_MUSTLOCK(s))
    SDL_UnlockSurface(s);
}

void image2matrix(SDL_Surface *image,
                               int Dx, int Dy, Matrice *M) {
  int i, j;
  
  M -> r = (image -> h / Dy);
  M -> c = (image -> w / Dx);
  
  if ((M -> m = (Uint8 *)calloc(M -> r * M -> c, sizeof(Uint8))) == NULL) {
    printf("Program exit\n");
    exit(1);
  }
  
  lockScreen(image);

  for (i = 0; i < M -> r; i++)
    for (j = 0; j < M -> c; j++)
      M -> m[Ind(i, j)] = getColor(image, j*Dx, i*Dy, M -> color);

  unlockScreen(image);
}

void multiColor(Uint8 v[], int colore) {
  if (colore <= 127) {
    v[R] = (colore * 255) / 127.5;
    v[G] = 255;
    v[B] = 0;
  }else {
    v[R] = 255;
    v[G] = 255 - (((colore - 127.5)*255)/127.5);
    v[B] = 0;
  }
}

void matrix2image(SDL_Surface *s, int scalax, int scalay, Matrice *M,
		  int multicolor) {
  int i, j;
  Uint8 v[3] = {0, 0, 0};
  
  lockScreen(s);

  for (i = 0; i < s -> h; i++)
    for (j = 0; j < s -> w; j++) {
      /*
      SDL_GetRGB(getpixel(s, i, j), s -> format, &v[R], &v[G], &v[B]); */
      if (multicolor)
	multiColor(v, M -> m[Ind((i/scalay), (j/scalax))]);
      else
	v[M -> color] = M -> m[Ind((i/scalay), (j/scalax))];
      
      putpixel(s, j, i,
               SDL_MapRGB(s -> format, v[R], v[G], v[B]));
    }
  unlockScreen(s);
}

SDL_Surface *loadImage(char *file) {
  SDL_Surface *image;

  if (file != NULL) {
    image = IMG_Load_RW(SDL_RWFromFile(file, "rb"), 1);
    
    background.x = background.y = 0;
    background.w = image -> w;
    background.h = image -> h;
    
  } else {
    image = SDL_CreateRGBSurface(SDL_HWSURFACE, background.w, background.h,
                                 BPP, rmask, gmask, bmask, amask);

    SDL_FillRect(image, &background,
                 SDL_MapRGB(image -> format, 255, 0, 0));
    
  }
  if (image == NULL) {
    fprintf(stderr, "Error: %s\n", IMG_GetError());
    exit(1);
  }
  return image;
}

/* D e' la matrice dei coefficienti di diffusione */
float **coefficient_matrix(int Dx, int Dy, double Dt, Matrice *D) {
  float **A;
  int k;
  int dim = D -> r * D -> c;

  if ((A = (float **)malloc(dim * sizeof(float *))) == NULL) exit(1);

  for (k = 0; k < dim; k++)
    if ((A[k] = (float *)calloc(dim, sizeof(float))) == NULL) {
      printf("the memory is below to run fick.\n");
      exit(1);
    }
    
  /* scrive i 5 elementi diversi da 0 per ogni riga di A */
  for (k = 0; k < dim; k++) {
    A[k][k] = 2; /* 2(1 + Lx + Ly) */

    if (k - 1 >= 0 && (k-1)%D->c != D->c - 1) {
      A[k][k-1] = -1 * Lx(k);  /* -Lx */
      A[k][k] += Lx(k);
    }
    if (k + 1 < dim && (k+1)%D->c != 0) {
      A[k][k+1] = -1 * Lx(k);  /* -Lx */
      A[k][k] += Lx(k);
    }
    if (k - D->c >= 0) {
      A[k][k - D->c] = -1 * Ly(k);  /* -Ly */
      A[k][k] += Ly(k);
    }
    if (k + D->c < dim) {
      A[k][k+D->c] = -1 * Ly(k);  /* -Ly */
      A[k][k] += Ly(k);
    }
  }
  return A;
}

/*  b[k] = 2(1-Lx-Ly)C(k) + LxC(k-1) + LxC(k+1) + LyC(k-N) + LyC(k+N)
 */
float *calculate_b(int Dx, int Dy, double Dt, Matrice *D, Matrice *C) {
  int k;
  float *b;

  if ((b = (float *)malloc(D -> r * D -> c * sizeof(float))) == NULL) exit(1);

  for (k = 0; k < (D -> r * D -> c); k++) {

    b[k] = (implicit+1)* C->m[k];           /* (2-2Lx-2Ly)C(k) implicit */
                                            /* (1-2Lx-2Ly)C(k) explicit */
    /*    printf("b[%d] = %f\n", k, b[k]); */
    if ((k-1)%D->c != D->c - 1 && k-1 >= 0) {
      b[k] += Lx(k) * C->m[k - 1];          /* LxC(k-1) */
      b[k] -= Lx(k) * C->m[k];
      /*      printf("k-1 b[%d] = %f\n", k, b[k]);*/
    }
    if ((k+1)%D->c != 0) {
      b[k] += Lx(k) * C->m[k + 1];          /* LxC(k+1) */
      b[k] -= Lx(k) * C->m[k];
      /*      printf("k+1 b[%d] = %f\n", k, b[k]);*/
    }
    if (k - D->c >= 0) {
      b[k] += Ly(k)* C->m[k - D->c];        /* LyC(k-N) */
      b[k] -= Ly(k) * C->m[k];
      /*      printf("k-N b[%d] = %f\n", k, b[k]); */
    }
    if (k + D->c < D -> r * D -> c) {
      b[k] += Ly(k)* C->m[k + D->c];      /* LyC(k+N) */
      b[k] -= Ly(k) * C->m[k];
      /*      printf("k+N b[%d] = %f\n", k, b[k]); */
    }
    /*    printf("fine\n", k, b[k]); */
  }
  return b;
}

void LUdecomposition(float **A, int dim) {
  int k, i, j;

  for (k = 0; k < dim; k++) {
    for (i = k + 1; i < dim; i++) {
      A[i][k] = A[i][k] / A[k][k];
    }
    for (i = k + 1; i < dim; i++)
      for (j = k + 1; j < dim; j++)
        A[i][j] = A[i][j] - A[i][k] * A[k][j];
  }
}

void LUsolve(float **LU, float *b, int dim, Uint8 *v) {
  int i, j;
  float *y, *vett, sum;

  if (((y = (float *)malloc(dim*sizeof(float))) == NULL) ||
      ((vett = (float *)malloc(dim*sizeof(float))) == NULL))
    exit(1);
  
  for (i = 0; i < dim; i++) {
    sum = 0;
    for(j = 0; j < i; j++)
      sum += LU[i][j] * y[j];
    y[i] = b[i] - sum;
  }

  for (i = dim - 1; i >= 0; i--) {
    sum = 0;
    for(j = i + 1; j < dim; j++)
      sum += LU[i][j] * vett[j];
    vett[i] = ((y[i] - sum) / LU[i][i]);

    if (((y[i] - sum) / LU[i][i]) > 255) {
      printf("%f\n", (y[i] - sum) / LU[i][i] );
      vett[i] = 255;
    }
    else vett[i] = ((y[i] - sum) / LU[i][i]) + 0.5;
    v[i] = (Uint8)vett[i];
  }
  free(y);
  free(vett);
}

void stop(SDL_Surface *screen, char *file, Uint8 *a, Uint8 *b, float *f[], int dim) {
  int i;
  if (file != NULL) {
    if (SDL_SaveBMP(screen, file) == -1)
      printf("error\n");
      }
  if (a != NULL) free(a);
  if (b != NULL) free(b);
  if (f != NULL) {
    for (i = 0; i < dim; i++)
      free(f[i]);
    free(f);
  }
  SDL_FreeSurface(screen);
  exit(0);
}
